# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is a mapping to a neo4j node for a file and filetype"""
import hashlib

import graphrepo.models.relationships as rel
from graphrepo.models.custom_node import CustomNode
from graphrepo.models.method import Method


class File(CustomNode):
    """File OGM Node - Maps files changed in a commit
    to py2neo objects
    """

    def __init__(self, file, project_id=None, graph=None, file_type=True):
        """Instantiates file object. If a graph is provided
        the node is also added to neo4j
        :param file: pydriller Modification object
        :param project_id: a string identifying the project a file belogns to
        :param graph: py2neo graph object
        :param file_type: flag which decides if the file type
          should be indexed
        """
        self.node_type = "File"
        self.node_index = "hash"
        self.project_id = project_id

        self.file = file
        att = self.get_update_attributes()
        _hash = hashlib.sha224(str(file.filename).encode('utf-8')).hexdigest()
        super().__init__(self.node_type, hash=_hash,
                         name=file.filename,
                         project_id=project_id,
                         **att)

        if graph is not None:
            self.index(graph)

        if file_type is True:
            self.index_type(graph=graph)

    def index_type(self, graph):
        """Creates file type if it does not exist and adds filetype
        relationship
        :param graph: py2neo Graph object
        """
        self.file_type = FileType(self.file, self.project_id, graph=graph)
        rel.FileType(self.file_type, self, graph=graph)

    def get_update_attributes(self):
        """Updates file attributes
        :returns: dictionary with custom node attributes
        """
        return {
            'old_path': self.file.old_path if self.file.old_path else '',
            'new_path': self.file.new_path if self.file.new_path else '',
            'source_code': self.file.source_code if self.file.source_code else '',
            'source_code_before':  self.file.source_code_before if self.file.source_code_before else '',
            'nloc': self.file.nloc if self.file.nloc else -1,
            'complexity': self.file.complexity if self.file.complexity else -1,
            'token_count': self.file.token_count if self.file.token_count else -1,
        }

    def index_methods(self, graph, commit=None, *args, **kwargs):
        """Indexes latest methods and adds a relation
        between commit and the methods changed in the commit
        if the commit object is given
        :param graph: py2neo Graph object
        :param commit: GraphRepo Commit object
        """
        # index new or changed methods
        for met in self.file.methods:
            method = Method(met, self.project_id, graph=graph)
            rel.HasMethod(rel_from=self, rel_to=method, graph=graph)
            if met in self.file.changed_methods:
                type_ = self._get_method_type(met)
                rel.UpdateMethod(rel_from=commit, rel_to=method,
                                 type=type_, graph=graph,
                                 *args, **kwargs)

        if len(self.file.methods):
            self.update_deleted_methods(graph, commit, *args, **kwargs)

    def update_deleted_methods(self, graph, commit, *args, **kwargs):
        """Indexes latest methods and adds a relation
        between commit and the methods changed in the commit
        if the commit object is given
        :param graph: py2neo Graph object
        :param commit: GraphRepo Commit object
        """
        for met in self.file.methods_before:
            if met not in self.file.methods:
                method = Method(met, self.project_id, graph=graph)
                # add relationship from commit to method
                rel.UpdateMethod(rel_from=commit, rel_to=method,
                                 type='DELETE', graph=graph, *args, **kwargs)
                # replace relationship from file to method from HasMethod to HadMethod
                has_rel = graph.match_one([self, method], rel.HasMethod)
                # TODO: Test this better
                if has_rel:
                    graph.separate(has_rel)
                rel.HadMethod(rel_from=self, rel_to=method, graph=graph)

    def _get_method_type(self, met):
        """Checks if the method existed before and it is updated"""
        type_ = "ADD"
        for m in self.file.methods_before:
            if m.name == met.name:
                return "MODIFY"
        return type_


class FileType(CustomNode):
    """Filetype OGM Node - Maps file types to py2neo object
    """

    def __init__(self, file, project_id=None, graph=None):
        """Instantiate Filetype object. If a graph is provided
        the node is also added to neo4j
        :param file: pydriller Modification object
        :param project_id: a string identifying the project a file belogns to
        :param graph: py2neo graph object
        """
        self.node_type = "FileType"
        self.node_index = "hash"

        _name = '.' + file.filename.split('.')[-1:][0]
        _hash = hashlib.sha224(_name.encode('utf-8')).hexdigest()
        super().__init__(self.node_type, hash=_hash, name=_name,
                         project_id=project_id)
        if graph is not None:
            self.index(graph)



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module is a custom node mapper, which translates
the neo4j node models to custom data structures, e.g. dictionaries """

from graphrepo.miners.mappers.default import DefaultMapper


class NodeMapper(DefaultMapper):
  """The miners are currently synchronous, but
  ideally they will be async in the future"""

  def __init__(self, *args, **kwargs):
    """Default Init"""
    super().__init__(*args, **kwargs)

  @classmethod
  def map_default_node(cls, node):
    """Maps a node to a dictionary"""
    return {
        'data': {
            'id': cls.get_id(node),
            'label': cls.get_label(node),
            'type': list(node._labels)[0],
            'name': node['name']
        }
    }

  @classmethod
  def get_id(cls, node):
    """Returns node default identifier"""
    return node.identity

  @classmethod
  def get_label(cls, node):
    """Returns node default label"""
    return node['name'] if node['name'] else node['hash']



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module establishes a connection to the database and
initializes the miners
"""
from py2neo import Graph
from graphrepo.config import Config
from graphrepo.logger import Logger
from graphrepo.singleton import Singleton
from graphrepo.miners import MineManager

LG = Logger()


class Miner(metaclass=Singleton):
  """Miner class - connects to db and initializes miners.
  The miners do the db queries.
  This class is a singleton because it holds the connection
  to Neo4j in self.graph
  """

  def __init__(self):
    """Initializes the properties of this class"""
    self.config = Config()
    self.graph = None
    self.manager = None

  def configure(self, db_url="localhost",
                port=13000, db_user="neo4j",
                db_pwd="neo4j", *args, **kwargs):
    """Sets the application constants"""
    # TODO: validate inputs
    self.config.DB_URL = db_url
    self.config.PORT = port
    self.config.DB_USER = db_user
    self.config.DB_PWD = db_pwd

    self.init_miners()

  def connect(self):
    """Instantiates the connection to Neo4j and stores
    the graph internally.
    Throws exception if the connection can not pe realized
    """
    try:
      self.graph = Graph(host=self.config.DB_URL,
                         user=self.config.DB_USER,
                         password=self.config.DB_PWD,
                         http_port=self.config.PORT)
    except Exception as exc:
      LG.log_and_raise(exc)

  def init_miners(self):
    """This method initializes the miner manager,
    which initializes a team of miners
    """
    self.config.check_config()
    self.check_connection()
    self.manager = MineManager(graph=self.graph)

  def check_connection(self):
    """Checks if there is a db connection and raises
    ReferenceError if not.
    """
    try:
      self.connect()
    except:
      raise ReferenceError("There is no valid "
                           "database connection. Please "
                           "configure and connect first.")



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is a mapping to a neo4j node for a year"""

from graphrepo.models.custom_node import CustomNode


class Year(CustomNode):
  """Year OGM node
  """

  def __init__(self, date, project_id=None, graph=None):
    """Instantiates a year object. If a graph is provided
    the object is indexed in neo4j
    :param date: datetime object containing all date info
    :param project_id: a string identifying the project a yeaer belogns to
    """
    self.node_type = "Year"
    self.node_index = "name"

    super().__init__(self.node_type, name=date.year, project_id=project_id)
    if graph is not None:
      self.index(graph)



from .branch import *
from .commit import *
from .custom_node import *
from .day import *
from .developer import *
from .file import *
from .method import *
from .month import *
from .relationships import *
from .year import *



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module is a custom relationship mapper, which translates
the neo4j relationship models to custom data structures, e.g. dictionaries """

from graphrepo.miners.mappers.default import DefaultMapper
from graphrepo.miners.mappers.node import NodeMapper


class RelMapper(DefaultMapper):
  """The miners are currently synchronous, but
  ideally they will be async in the future"""

  def __init__(self, *args, **kwargs):
    """Default Init"""
    super().__init__(*args, **kwargs)

  @classmethod
  def map_default_rel(cls, rel):
    """Maps a rellationship to a dictionary"""
    # Neo4j sometimes returns the labels in reverse order
    # so we concatenate the labels
    if len(rel.labels) < 2:
      label2 = list(rel.labels)[0]
    else:
      label2 = list(rel.labels)[0] + list(rel.labels)[1]

    return {
        'data': {
            'source': NodeMapper.get_id(rel.start_node),
            'target': NodeMapper.get_id(rel.end_node),
            'label': list(rel.labels)[0],
            'label_concat': label2,
            'id': rel.identity
        }
    }



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is a mapping between Pydriller Commit and neo4j """
import graphrepo.models.relationships as rel

from graphrepo import models as mdl


class Commit(mdl.CustomNode):
    """Commit OGM  - Mapps Commit from PyDriller to py2neo
    """

    def __init__(self, commit, config, graph=None, repo=None):
        """Instantiates a commit. If a graph is given
        the node is created in the graph
        :param commit: pydriller Commit object
        :param config: a Config object
        :param graph: py2neo Graph object
        :param repo: pydriller RepositoryMining object
        """
        self.node_type = "Commit"
        self.node_index = "hash"
        self.config = config
        self.project_id = config.PROJECT_ID

        self.commit = commit
        self.indexed = False

        self.author_dates = {
            'author_datetime': self.commit.author_date.strftime(
                "%Y/%m/%d, %H:%M:%S"),
            'author_date':  self.commit.author_date.strftime("%Y/%m/%d"),
            'author_time': self.commit.author_date.strftime("%H:%M")

        }
        mdl.CustomNode.__init__(self, self.node_type,
                                hash=self.commit.hash,
                                project_id=self.project_id,
                                ** self.author_dates)

        if graph is not None:
            self.index_all_data(graph=graph, repo=repo)

    def update_attributes(self, graph):
        """Updates custom attributes for commit node
        :param graph: py2neo graph object
        """
        self['is_merge'] = 1 if self.commit.merge else 0
        self['dmm_unit_size'] = self.commit.dmm_unit_size if self.commit.dmm_unit_size else -1
        self['dmm_unit_complexity'] = self.commit.dmm_unit_complexity if self.commit.dmm_unit_complexity else -1
        self['dmm_unit_interfacing'] = self.commit.dmm_unit_interfacing if self.commit.dmm_unit_interfacing else -1

        graph.push(self)

    def index_files_changed(self, graph):
        """Indexes the files changed by a commit
        :param graph: py2neo graph object
        """
        self.check_self(graph)
        for chg in self.commit.modifications:
            # Whenever a commit touches a method, the method
            # attributes are updated (e.g. loc or complexity)
            change = mdl.File(chg, self.project_id, graph=graph)
            rel.UpdateFile(rel_from=self, rel_to=change,
                           graph=graph, **self.author_dates)
            # index file methods and add relationship to change
            # methods in this commit
            change.index_methods(graph, self, **self.author_dates)

    def index_author(self, graph):
        """Indexes the commit author node in the graph
        and adds relationship to this commit
        :params graph: py2neo graph object
        """
        self.check_self(graph)
        dev = mdl.Developer(self.commit.author, graph=graph)
        rel.Authorship(rel_from=dev, rel_to=self, graph=graph)

    def index_parents(self, graph, repo):
        """This method chooses between indexing a parent
        relation or indexing a branch relationship between
        parent commits.
        :param graph: py2neo graph
        """
        if self.config.BRANCH_AS_NODE is True:
            self.index_parent(graph, repo)
        else:
            self.index_parent_branch(graph, repo)

    def index_parent(self, graph=None, repo=None):
        """For each parent of the commit, this method requests
        a commit object from pydriller's RepositoryMining object
        and indexes some data. If index_all_data is used, this method
        is applied recursively to all parents, however, this is
        very slow.
        :param graph: py2neo graph object
        :param repo: PyDriller RepositoryMining object
        """
        self.check_self(graph)
        for parent in self.commit.parents:
            commit = mdl.Commit(repo.get_commit(parent), self.config)
            commit.index_author(graph)
            commit.index_date(graph)
            # commit.index_all_data(graph, repo)
            rel.Parent(rel_from=commit, rel_to=self, graph=graph)
            # index branches
            branches = self.commit.branches
            for branch in branches:
                brn = mdl.Branch(
                    name=branch, project_id=self.project_id, graph=graph)
                rel.BelongsToBranch(rel_from=self, rel_to=brn,
                                    graph=graph, name=branch)

    def index_parent_branch(self, graph, repo):
        """For each parent of the commit, this method requests
        a commit object from pydriller's RepositoryMining object
        and indexes a branch relationship between the two and
        the parent commit data
        :param graph: py2neo Graph object
        :param repo: Pydriller RepositoryMining object
        """
        self.check_self(graph)
        for parent in self.commit.parents:
            commit = mdl.Commit(repo.get_commit(parent), self.config)
            commit.index_author(graph)
            commit.index_date(graph)
            common_branches = self._common_branch(commit)
            for branch in common_branches:
                rel.BelongsToBranch(rel_from=commit, rel_to=self,
                                    graph=graph, name=branch)

    def index_date(self, graph):
        """Splits the date in year, month and day and creates
        the graph nodes and relationships
        :param graph: py2neo graph object
        """
        self.check_self(graph)

        date = self.commit.author_date
        day = mdl.Day(date, self.project_id,  graph=graph)
        month = mdl.Month(date, self.project_id, graph=graph)
        year = mdl.Year(date, self.project_id, graph=graph)

        rel.YearMonth(rel_from=year, rel_to=month, graph=graph)
        rel.YearMonth(rel_from=year, rel_to=month, graph=graph)
        rel.MonthDay(rel_from=month, rel_to=day, graph=graph)
        rel.DayCommit(rel_from=day, rel_to=self, graph=graph)

    def index_all_data(self, graph, repo=None):
        """Indexes all the data for a commit
        :param graph: py2neo graph object
        :param repo: PyDriller RepositoryMining class for the
          current repo
        """
        self.update_attributes(graph=graph)
        self.check_self(graph)
        self.index_author(graph)
        self.index_date(graph)
        self.index_files_changed(graph)

        if repo is not None:
            self.index_parents(graph, repo)

    def _common_branch(self, ot_commit):
        """Searches for a common branch between one commit
        and self
        :param ot_commit: another commit
        """
        oth_branches = ot_commit.commit.branches
        curr_branches = self.commit.branches
        return list((oth_branches).intersection(curr_branches))



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module represents a custom py2neo Node which
should be inherited by all models
"""
from py2neo.data import Node


class CustomNode(Node):
    """Parent class for all custom Nodes"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.indexed = False

    def check_self(self, graph):
        """Checks if the node corresponding to this instance
        is indexed in the database
        :param graph: py2neo graph object
        """
        if not self.indexed:
            graph.merge(self, self.node_type, self.node_index)
            self.indexed = True

    def index(self, graph):
        """Adds graph node for this object
        :param graph: py2neo graph
        """
        graph.merge(self, self.node_type, self.node_index)



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml

from datetime import datetime


def parse_config(path):
    with open(path, 'r') as ymlfile:
        conf = yaml.load(ymlfile, Loader=yaml.FullLoader)

    neo = conf['neo']
    project = conf['project']

    project['start_date'] = datetime.strptime(
        project['start_date'], '%d %B, %Y %H:%M') if project['start_date'] else None
    project['end_date'] = datetime.strptime(
        project['end_date'], '%d %B, %Y %H:%M') if project['end_date'] else None

    return neo, project



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is a mapping to a neo4j node for a day"""
import hashlib

from graphrepo.models.custom_node import CustomNode


class Day(CustomNode):
  """Day node
  """

  def __init__(self, date, project_id=None, graph=None):
    """Instantiates a day object and hashes the month, year and
    day as an unique key for the day. We wish to simulate a time
    series (otherwise, each month could've connect to the same day).
    If a graph is provided, the object is indexed in neo4j
    :param date: datetime object containing all date info
    :param project_id: a string identifying the project a day belogns to
    """
    self.node_type = "Day"
    self.node_index = "hash"

    _hash = hashlib.sha224(
        (str(date.month) + str(date.year)
         + str(date.day)).encode('utf-8')).hexdigest()
    super().__init__(self.node_type, name=date.day, hash=_hash,
                     project_id=project_id)
    if graph is not None:
      self.index(graph)



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module stores all config constants. It is a singleton
because it is used across in several modules inside the app"""

from graphrepo.singleton import Singleton


class Config(metaclass=Singleton):
  """This class contains all config flags"""
  DB_URL = ""
  PORT = 0
  DB_USER = ""
  DB_PWD = ""
  REPO = ""
  START_DATE = None
  END_DATE = None
  PROJECT_ID = None
  # If True, each branch will be indexed as a node
  # and commits will be linked by a Parent relationship
  # If False, then the commits are linked by a Branch
  # relationship
  BRANCH_AS_NODE = True

  def check_config(self):
    """Checks if the config properties are set and
    raises ValueError if any value misses"""

    if self.DB_URL == "":
      raise ValueError("Database URL is not set.")

    if self.PORT == 0:
      raise ValueError("Database port is not set.")

    if self.DB_USER == "" or self.DB_PWD == "":
      raise ValueError("Database credentials are not set.")



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" This module uses pydriller to search a repository
and indexes it in neo4j
"""
from py2neo import Graph
from pydriller import RepositoryMining, GitRepository
from graphrepo.config import Config
from graphrepo.logger import Logger
from graphrepo.models.commit import Commit
from graphrepo.singleton import Singleton


LG = Logger()


class Driller(metaclass=Singleton):
    """Drill class - parses a git repo and uses the models
    to index everything in Neo4j. This class is a singleton
    because it holds the connection to Neo4j in self.graph
    """

    def __init__(self):
        """Initializes the properties of this class"""
        self.config = Config()
        self.graph = None

    def configure(self, db_url="localhost",
                  port=13000, db_user="neo4j",
                  db_pwd="neo4j", repo="pydriller/",
                  start_date=None,
                  end_date=None,
                  project_id=None):
        """Sets the application constants"""
        # TODO: validate inputs
        self.config.DB_URL = db_url
        self.config.PORT = port
        self.config.DB_USER = db_user
        self.config.DB_PWD = db_pwd
        self.config.REPO = repo
        self.config.START_DATE = start_date
        self.config.END_DATE = end_date
        self.config.PROJECT_ID = project_id

    def connect(self):
        """Instantiates the connection to Neo4j and stores
        the graph internally.
        Throws exception if the connection can not pe realized
        """
        try:
            self.graph = Graph(host=self.config.DB_URL,
                               user=self.config.DB_USER,
                               password=self.config.DB_PWD,
                               http_port=self.config.PORT)
        except Exception as exc:
            LG.log_and_raise(exc)

    def _drill(self):
        """Parses all commits
        :returns: a tupple containing a list of commits and
          a Pydriller GitRepository commit
        """
        rep_obj = GitRepository(self.config.REPO)
        commits = []
        for commit in RepositoryMining(self.config.REPO,
                                       since=self.config.START_DATE,
                                       to=self.config.END_DATE).traverse_commits():
            commits.append(Commit(commit, self.config))

        return commits, rep_obj

    def drill(self):
        """Gets commits and indexes them to Neo4j Database"""
        try:
            self.config.check_config()
            self.check_connection()

            commits, rep_obj = self._drill()
            for com in commits:
                com.index_all_data(self.graph, rep_obj)
        except Exception as exc:
            LG.log_and_raise(exc)
        else:
            return

    def check_connection(self):
        """Checks if there is a db connection and raises
        ReferenceError if not.
        """
        try:
            self.connect()
        except:
            raise ReferenceError("There is no valid "
                                 "database connection. Please "
                                 "configure and connect first.")

    def clean(self):
        """Removes all data in a graph
        """
        try:
            self.config.check_config()
            self.check_connection()

            self.graph.run("MATCH (n) OPTIONAL MATCH (n)-[r]-() DELETE n,r")
        except Exception as exc:
            LG.log_and_raise(exc)



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from py2neo import Graph, NodeMatcher, RelationshipMatcher
from graphrepo.config import Config
from graphrepo.logger import Logger
from graphrepo.singleton import Singleton
from graphrepo import miners
from graphrepo.miners import mappers


LG = Logger()


class MineManager():
    """MineManageer class - This class manages custom
    miners. At the moment we instantiate all miners,
    but other managers which handle different 'teams of miners'
    can be created.
    """

    def __init__(self, graph):
        """Initializes the properties of this class"""
        self.graph = graph
        self.node_matcher = NodeMatcher(graph)
        self.rel_matcher = RelationshipMatcher(graph)
        self.init_miners()

    def init_miners(self):
        """Initializes all miners"""
        try:
            # TODO: Parse this automatically?
            self.commit_miner = miners.CommitMiner(
                graph=self.graph,
                node_matcher=self.node_matcher,
                rel_matcher=self.rel_matcher)
            self.dev_miner = miners.DeveloperMiner(graph=self.graph,
                                                   node_matcher=self.node_matcher,
                                                   rel_matcher=self.rel_matcher)
            self.file_miner = miners.FileMiner(graph=self.graph,
                                               node_matcher=self.node_matcher,
                                               rel_matcher=self.rel_matcher)

        except Exception as exc:
            LG.log_and_raise(exc)
        else:
            return

    def get_all_data(self, map=False, merge=False):
        """Returns all nodes and relationships from Neo4j
        :param map: Map data using default mappers for nodes and rels
        :param merge: Concatenates nodes an relationships
        :returns: if merge is False it returns a tuple with two
          arrays: the first with nodes, the second with relationships
          if merge is True it returns only one array
        """
        nodes = self.node_matcher.match()
        rels = self.rel_matcher.match()

        if map is True:
            nodes = [mappers.node.NodeMapper.map_default_node(
                n) for n in nodes]
            rels = [mappers.rels.RelMapper.map_default_rel(r) for r in rels]
            if merge is True:
                return nodes+rels, None

        return nodes, rels

    def remove_all_data(self, project_id):
        """Removes all elements from a graph given a project id
        :param project_id: the project ids for the nodes to be removed
        """
        pass
        # self.graph.run()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module mines commits and contains all related Neo4j queries"""

from graphrepo.miners.default import DefaultMiner


class CommitMiner(DefaultMiner):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def query(self, **kwargs):
    return self.node_matcher.match("Commit", **kwargs)

  def get_all(self):
    return self.node_matcher.match("Commit")



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is a mapping to a neo4j node for a day"""

from graphrepo.models.custom_node import CustomNode


class Branch(CustomNode):
  """Branch node
  """

  def __init__(self, name, project_id=None, graph=None, *args, **kwargs):
    """Instantiates a branch node and, if a graph is given,
    it indexes the graph to Neo4j
    :param name: string containing an unique name
    :param project_id: a string identifying the project a branch belogns to
    :param graph: py2neo graph objects
    """
    self.node_type = "Branch"
    self.node_index = "name"

    super().__init__(self.node_type, name=name,
                     project_id=project_id, *args, **kwargs)
    if graph is not None:
      self.index(graph)



from setuptools import setup, find_packages

setup(name="graphrepo",
      version="0.1.4",
      description="A tool that maps a Github repo to Neo4j and Helps Mining the Repo in the DB",
      url="https://github.com/NullConvergence/GraphRepo",
      license='Apache License',
      python_requires='>=3.5',
      install_requires=[
          'py2neo==4.2.0',
          'pydriller==1.7'
      ],
      packages=find_packages(),
      package_dir={'graphrepo': 'graphrepo'})



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module holds all possible relationships in the graph"""
from py2neo.data import Relationship


class CustomRelationship(Relationship):
    """Extends py2neo relationship object
    with custom functionality.
    """

    def __init__(self, rel_from=None, rel_to=None, graph=None, *args, **kwargs):
        """Instantiates a custom relationship. If a graph is given
        the relationship is indexed in the graph.
        :param rel_from: py2neo Node object
        :param rel_to: py2neo Node object
        :param graph: py2neo graph object
        """
        super().__init__(rel_from, rel_to, *args, **kwargs)
        if graph is not None:
            self.create(graph)

    def create(self, graph):
        graph.create(self)


class BelongsToBranch(CustomRelationship):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Authorship(CustomRelationship):
    pass


class Parent(CustomRelationship):
    pass


class YearMonth(CustomRelationship):
    pass


class MonthDay(CustomRelationship):
    pass


class DayCommit(CustomRelationship):
    pass


class UpdateFile(CustomRelationship):
    def __init__(self, rel_from, rel_to, graph, *args, **kwargs):
        """Instantiates an Update relationship and creates a Neo4j relationship
        which also contains some module metrics extracted from Pydriller through
        Lizard.
        :param rel_from: Commit object
        :param rel_to: File object
        :param graph: Py2neo Graph object
        """
        metrics = self.get_metrics(rel_to)
        attributes = rel_to.get_update_attributes()
        super().__init__(rel_from, rel_to, graph,
                         *args, **kwargs,
                         **metrics,  **attributes)

    def get_metrics(self, change):
        """Creates and returns a dic with different metrics for a file
        :param commit: Commit object
        :returns: dic
        """
        return {
            'added': change.file.added,
            'removed': change.file.removed,
            'type': change.file.change_type.name,
        }


class HasMethod(CustomRelationship):
    def __init__(self, rel_from, rel_to, graph, *args, **kwargs):
        """Instantiates a Method relationship and creates a Neo4j relationship
        :param rel_from: Update object
        :param rel_to: Method Node
        :param graph: Py2neo Graph
        """
        super().__init__(rel_from, rel_to, graph, *args, **kwargs)


class HadMethod(CustomRelationship):
    def __init__(self, rel_from, rel_to, graph, *args, **kwargs):
        """Instantiates a HadMethod relationship and creates a Neo4j relationship
        :param rel_from: Update object
        :param rel_to: Method Node
        :param graph: Py2neo Graph
        """
        super().__init__(rel_from, rel_to, graph, *args, **kwargs)


class UpdateMethod(CustomRelationship):
    def __init__(self, rel_from, rel_to, graph, type='ADD', *args, **kwargs):
        """Instantiates an UpdateMethod relationship and creates a Neo4j relationship
        :param rel_from: Update object
        :param rel_to: Method Node
        :param graph: Py2neo Graph
        """
        method_metrics = rel_to.get_metrics()
        super().__init__(rel_from, rel_to, graph, type=type,
                         *args, **kwargs,
                         **method_metrics)


class FileType(CustomRelationship):
    pass



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is a mapping to a neo4j node for a method"""

from graphrepo.models.custom_node import CustomNode


class Method(CustomNode):
    """Method node
    """

    def __init__(self, method, project_id=None, graph=None):
        """Instantiates a method object and creates an unique
        hash
        :param method: PyDriller Method object
        :param project_id: a string identifying the project a file belogns to
        """
        self.node_type = "Method"
        self.node_index = "name"

        self.method = method
        self.metrics = self.get_metrics()

        super().__init__(self.node_type, name=self.method.name,
                         file_name=self.method.filename,
                         long_name=self.method.long_name,
                         project_id=project_id, **self.metrics)
        if graph is not None:
            self.index(graph)

    def get_metrics(self):
        """Creates and returns a dic with different metrics for a method
        :returns: dic
        """
        return {
            'parameters': self.method.parameters,
            'complexity': self.method.complexity,
            'nloc': self.method.nloc,
            'fan_in': self.method.fan_in,
            'fan_out': self.method.fan_out,
            'general_fan_out': self.method.general_fan_out,
            'length': self.method.length,
            'token_count': self.method.token_count,
            'start_line': self.method.start_line,
            'end_line': self.method.end_line
        }



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import yaml
import graphrepo.models.relationships as rel

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.driller import Driller
from graphrepo.utils import parse_config
from graphrepo import models as mdl


class TestCommit:
    """Most data is indexed when indexing a commmit
    so this class tests indexing for multiple models"""

    def test_nodes_index(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        neo, project = parse_config(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller = Driller()

        test_driller.configure(**neo, **project)
        test_driller.connect()

        test_driller.drill()
        # test that all nodes were indexed
        node_matcher = NodeMatcher(test_driller.graph)
        all_commits = list(node_matcher.match("Commit"))
        assert len(all_commits) == 8

        all_devs = list(node_matcher.match("Developer"))
        assert len(all_devs) == 2

        all_files = list(node_matcher.match("File"))
        assert len(all_files) == 5

        all_file_types = list(node_matcher.match("  "))
        assert len(all_file_types) == 3

        all_methods = list(node_matcher.match("Method"))
        assert len(all_methods) == 2

        all_days = list(node_matcher.match("Day"))
        assert len(all_days) == 2

        all_months = list(node_matcher.match("Month"))
        assert len(all_months) == 1

        all_years = list(node_matcher.match("Year"))
        assert len(all_years) == 1

        test_driller.clean()

    def test_rel_index(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        neo, project = parse_config(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller = Driller()

        test_driller.configure(**neo, **project)
        test_driller.connect()

        test_driller.drill()

        # test that all relationships were indexed
        rel_matcher = RelationshipMatcher(test_driller.graph)

        all_branch = list(rel_matcher.match(None, "BelongsToBranch"))
        assert len(all_branch) == 7

        all_authorship = list(rel_matcher.match(None, "Authorship"))
        assert len(all_authorship) == 8

        all_parent = list(rel_matcher.match(None, "Parent"))
        assert len(all_parent) == 8

        all_ym = list(rel_matcher.match(None, "YearMonth"))
        assert len(all_ym) == 1

        all_md = list(rel_matcher.match(None, "MonthDay"))
        assert len(all_md) == 2

        all_dc = list(rel_matcher.match(None, "DayCommit"))
        assert len(all_dc) == 8

        all_updadedfile = list(rel_matcher.match(None, "UpdateFile"))
        assert len(all_updadedfile) == 9

        all_hasmethod = list(rel_matcher.match(None, "HasMethod"))
        assert len(all_hasmethod) == 2

        all_hadmethod = list(rel_matcher.match(None, "HadMethod"))
        assert len(all_hadmethod) == 1

        all_updatemethod = list(rel_matcher.match(None, "UpdateMethod"))
        assert len(all_updatemethod) == 6

        all_filetype = list(rel_matcher.match(None, "FileType"))
        assert len(all_filetype) == 5

        test_driller.clean()

    def test_custom_attributes_rel(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        neo, project = parse_config(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller = Driller()

        test_driller.configure(**neo, **project)
        test_driller.connect()

        test_driller.drill()

        node_matcher = NodeMatcher(test_driller.graph)
        rel_matcher = RelationshipMatcher(test_driller.graph)

        commit = node_matcher.match(
            "Commit", hash="7a0d1eac6dd6a7af7ff0b0f4927b9beaedaacd52").first()
        assert commit['dmm_unit_complexity'] == 1

        update_file_rel = rel_matcher.match([commit], "UpdateFile").first()
        assert update_file_rel['complexity'] == 2
        assert update_file_rel['nloc'] == 8
        assert update_file_rel['old_path'] == 'gr_test/default_class.py'
        assert update_file_rel['new_path'] == 'gr_test/default_class.py'
        assert update_file_rel['token_count'] == 42
        assert update_file_rel['type'] == 'MODIFY'
        assert update_file_rel['removed'] == 6
        assert update_file_rel['added'] == 0

        update_method_rel = rel_matcher.match(
            [commit], 'UpdateMethod').first()
        assert update_method_rel['type'] == 'DELETE'
        assert update_method_rel['nloc'] == 5
        assert update_method_rel['complexity'] == 2
        assert update_method_rel['token_count'] == 21
        assert update_method_rel['length'] == 5
        assert update_method_rel['fan_in'] == 0
        assert update_method_rel['fan_out'] == 0
        assert update_method_rel['start_line'] == 11
        assert update_method_rel['end_line'] == 15
